package xyz.lychee.lagfixer.libs.adventure.platform.facet;

import java.io.Closeable;
import java.util.Collection;
import java.util.Collections;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;
import xyz.lychee.lagfixer.libs.adventure.audience.Audience;
import xyz.lychee.lagfixer.libs.adventure.audience.MessageType;
import xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar;
import xyz.lychee.lagfixer.libs.adventure.chat.ChatType;
import xyz.lychee.lagfixer.libs.adventure.chat.SignedMessage;
import xyz.lychee.lagfixer.libs.adventure.identity.Identity;
import xyz.lychee.lagfixer.libs.adventure.inventory.Book;
import xyz.lychee.lagfixer.libs.adventure.pointer.Pointers;
import xyz.lychee.lagfixer.libs.adventure.sound.Sound;
import xyz.lychee.lagfixer.libs.adventure.sound.SoundStop;
import xyz.lychee.lagfixer.libs.adventure.sound.Sound.Emitter;
import xyz.lychee.lagfixer.libs.adventure.text.Component;
import xyz.lychee.lagfixer.libs.adventure.text.flattener.ComponentFlattener;
import xyz.lychee.lagfixer.libs.adventure.title.Title;
import xyz.lychee.lagfixer.libs.adventure.title.TitlePart;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;
import xyz.lychee.lagfixer.libs.annotations.ApiStatus.Internal;
import xyz.lychee.lagfixer.libs.annotations.ApiStatus.OverrideOnly;

@Internal
public class FacetAudience<V> implements Audience, Closeable {
   protected final @NotNull FacetAudienceProvider<V, FacetAudience<V>> provider;
   private final @NotNull Set<V> viewers;
   private @Nullable V viewer;
   private volatile Pointers pointers;
   private final Facet.@Nullable Chat<V, Object> chat;
   private final Facet.@Nullable ActionBar<V, Object> actionBar;
   private final Facet.@Nullable Title<V, Object, Object, Object> title;
   private final Facet.@Nullable Sound<V, Object> sound;
   private final Facet.@Nullable EntitySound<V, Object> entitySound;
   private final Facet.@Nullable Book<V, Object, Object> book;
   private final Facet.BossBar.@Nullable Builder<V, Facet.BossBar<V>> bossBar;
   private final @Nullable Map<BossBar, Facet.BossBar<V>> bossBars;
   private final Facet.@Nullable TabList<V, Object> tabList;
   private final @NotNull Collection<? extends Facet.Pointers<V>> pointerProviders;

   public FacetAudience(final @NotNull FacetAudienceProvider provider, final @NotNull Collection<? extends V> viewers, final @Nullable Collection<? extends Facet.Chat> chat, final @Nullable Collection<? extends Facet.ActionBar> actionBar, final @Nullable Collection<? extends Facet.Title> title, final @Nullable Collection<? extends Facet.Sound> sound, final @Nullable Collection<? extends Facet.EntitySound> entitySound, final @Nullable Collection<? extends Facet.Book> book, final @Nullable Collection<? extends Facet.BossBar.Builder> bossBar, final @Nullable Collection<? extends Facet.TabList> tabList, final @Nullable Collection<? extends Facet.Pointers> pointerProviders) {
      this.provider = (FacetAudienceProvider)Objects.requireNonNull(provider, "audience provider");
      this.viewers = new CopyOnWriteArraySet();
      Iterator var12 = ((Collection)Objects.requireNonNull(viewers, "viewers")).iterator();

      while(var12.hasNext()) {
         V viewer = var12.next();
         this.addViewer(viewer);
      }

      this.refresh();
      this.chat = (Facet.Chat)Facet.of(chat, this.viewer);
      this.actionBar = (Facet.ActionBar)Facet.of(actionBar, this.viewer);
      this.title = (Facet.Title)Facet.of(title, this.viewer);
      this.sound = (Facet.Sound)Facet.of(sound, this.viewer);
      this.entitySound = (Facet.EntitySound)Facet.of(entitySound, this.viewer);
      this.book = (Facet.Book)Facet.of(book, this.viewer);
      this.bossBar = (Facet.BossBar.Builder)Facet.of(bossBar, this.viewer);
      this.bossBars = this.bossBar == null ? null : Collections.synchronizedMap(new IdentityHashMap(4));
      this.tabList = (Facet.TabList)Facet.of(tabList, this.viewer);
      this.pointerProviders = (Collection)(pointerProviders == null ? Collections.emptyList() : pointerProviders);
   }

   public void addViewer(final @NotNull V viewer) {
      if (this.viewers.add(viewer) && this.viewer == null) {
         this.viewer = viewer;
         this.refresh();
      }

   }

   public void removeViewer(final @NotNull V viewer) {
      if (this.viewers.remove(viewer) && this.viewer == viewer) {
         this.viewer = this.viewers.isEmpty() ? null : this.viewers.iterator().next();
         this.refresh();
      }

      if (this.bossBars != null) {
         Iterator var2 = this.bossBars.values().iterator();

         while(var2.hasNext()) {
            Facet.BossBar<V> listener = (Facet.BossBar)var2.next();
            listener.removeViewer(viewer);
         }

      }
   }

   public void refresh() {
      synchronized(this) {
         this.pointers = null;
      }

      if (this.bossBars != null) {
         Iterator var1 = this.bossBars.entrySet().iterator();

         while(var1.hasNext()) {
            Map.Entry<BossBar, Facet.BossBar<V>> entry = (Map.Entry)var1.next();
            BossBar bar = (BossBar)entry.getKey();
            Facet.BossBar<V> listener = (Facet.BossBar)entry.getValue();
            listener.bossBarNameChanged(bar, bar.name(), bar.name());
         }

      }
   }

   public void sendMessage(final @NotNull Identity source, final @NotNull Component original, final @NotNull MessageType type) {
      if (this.chat != null) {
         Object message = this.createMessage(original, this.chat);
         if (message != null) {
            Iterator var5 = this.viewers.iterator();

            while(var5.hasNext()) {
               V viewer = var5.next();
               this.chat.sendMessage(viewer, source, message, type);
            }

         }
      }
   }

   public void sendMessage(final @NotNull Component original, final ChatType.@NotNull Bound boundChatType) {
      if (this.chat != null) {
         Object message = this.createMessage(original, this.chat);
         if (message != null) {
            Component name = this.provider.componentRenderer.render(boundChatType.name(), this);
            Component target = null;
            if (boundChatType.target() != null) {
               target = this.provider.componentRenderer.render(boundChatType.target(), this);
            }

            Object renderedType = boundChatType.type().bind(name, target);
            Iterator var7 = this.viewers.iterator();

            while(var7.hasNext()) {
               V viewer = var7.next();
               this.chat.sendMessage(viewer, Identity.nil(), message, renderedType);
            }

         }
      }
   }

   public void sendMessage(final @NotNull SignedMessage signedMessage, final ChatType.@NotNull Bound boundChatType) {
      if (signedMessage.isSystem()) {
         Component content = signedMessage.unsignedContent() != null ? signedMessage.unsignedContent() : Component.text(signedMessage.message());
         this.sendMessage((Component)content, boundChatType);
      } else {
         super.sendMessage(signedMessage, boundChatType);
      }

   }

   public void sendActionBar(final @NotNull Component original) {
      if (this.actionBar != null) {
         Object message = this.createMessage(original, this.actionBar);
         if (message != null) {
            Iterator var3 = this.viewers.iterator();

            while(var3.hasNext()) {
               V viewer = var3.next();
               this.actionBar.sendMessage(viewer, message);
            }

         }
      }
   }

   public void playSound(final @NotNull Sound original) {
      if (this.sound != null) {
         Iterator var2 = this.viewers.iterator();

         while(var2.hasNext()) {
            V viewer = var2.next();
            Object position = this.sound.createPosition(viewer);
            if (position != null) {
               this.sound.playSound(viewer, original, position);
            }
         }

      }
   }

   public void playSound(final @NotNull Sound sound, final Sound.@NotNull Emitter emitter) {
      if (this.entitySound != null) {
         Object message;
         if (emitter == Emitter.self()) {
            Iterator var3 = this.viewers.iterator();

            while(var3.hasNext()) {
               V viewer = var3.next();
               message = this.entitySound.createForSelf(viewer, sound);
               if (message != null) {
                  this.entitySound.playSound(viewer, message);
               }
            }
         } else {
            Object message = this.entitySound.createForEmitter(sound, emitter);
            if (message == null) {
               return;
            }

            Iterator var7 = this.viewers.iterator();

            while(var7.hasNext()) {
               message = var7.next();
               this.entitySound.playSound(message, message);
            }
         }

      }
   }

   public void playSound(final @NotNull Sound original, final double x, final double y, final double z) {
      if (this.sound != null) {
         Object position = this.sound.createPosition(x, y, z);
         Iterator var9 = this.viewers.iterator();

         while(var9.hasNext()) {
            V viewer = var9.next();
            this.sound.playSound(viewer, original, position);
         }

      }
   }

   public void stopSound(final @NotNull SoundStop original) {
      if (this.sound != null) {
         Iterator var2 = this.viewers.iterator();

         while(var2.hasNext()) {
            V viewer = var2.next();
            this.sound.stopSound(viewer, original);
         }

      }
   }

   public void openBook(final @NotNull Book original) {
      if (this.book != null) {
         String title = this.toPlain(original.title());
         String author = this.toPlain(original.author());
         List<Object> pages = new LinkedList();
         Iterator var5 = original.pages().iterator();

         Object viewer;
         while(var5.hasNext()) {
            Component originalPage = (Component)var5.next();
            viewer = this.createMessage(originalPage, this.book);
            if (viewer != null) {
               pages.add(viewer);
            }
         }

         if (title != null && author != null && !pages.isEmpty()) {
            Object book = this.book.createBook(title, author, pages);
            if (book != null) {
               Iterator var9 = this.viewers.iterator();

               while(var9.hasNext()) {
                  viewer = var9.next();
                  this.book.openBook(viewer, book);
               }

            }
         }
      }
   }

   private String toPlain(final Component comp) {
      if (comp == null) {
         return null;
      } else {
         StringBuilder builder = new StringBuilder();
         ComponentFlattener var10000 = ComponentFlattener.basic();
         Component var10001 = this.provider.componentRenderer.render(comp, this);
         Objects.requireNonNull(builder);
         var10000.flatten(var10001, builder::append);
         return builder.toString();
      }
   }

   public void showTitle(final Title original) {
      if (this.title != null) {
         Object mainTitle = this.createMessage(original.title(), this.title);
         Object subTitle = this.createMessage(original.subtitle(), this.title);
         Title.Times times = original.times();
         int inTicks = times == null ? -1 : this.title.toTicks(times.fadeIn());
         int stayTicks = times == null ? -1 : this.title.toTicks(times.stay());
         int outTicks = times == null ? -1 : this.title.toTicks(times.fadeOut());
         Object collection = this.title.createTitleCollection();
         if (inTicks != -1 || stayTicks != -1 || outTicks != -1) {
            this.title.contributeTimes(collection, inTicks, stayTicks, outTicks);
         }

         this.title.contributeSubtitle(collection, subTitle);
         this.title.contributeTitle(collection, mainTitle);
         Object title = this.title.completeTitle(collection);
         if (title != null) {
            Iterator var10 = this.viewers.iterator();

            while(var10.hasNext()) {
               V viewer = var10.next();
               this.title.showTitle(viewer, title);
            }

         }
      }
   }

   public <T> void sendTitlePart(@NotNull final TitlePart<T> part, @NotNull final T value) {
      if (this.title != null) {
         Objects.requireNonNull(value, "value");
         Object collection = this.title.createTitleCollection();
         Object title;
         if (part == TitlePart.TITLE) {
            title = this.createMessage((Component)value, this.title);
            if (title != null) {
               this.title.contributeTitle(collection, title);
            }
         } else if (part == TitlePart.SUBTITLE) {
            title = this.createMessage((Component)value, this.title);
            if (title != null) {
               this.title.contributeSubtitle(collection, title);
            }
         } else {
            if (part != TitlePart.TIMES) {
               throw new IllegalArgumentException("Unknown TitlePart '" + part + "'");
            }

            Title.Times times = (Title.Times)value;
            int inTicks = this.title.toTicks(times.fadeIn());
            int stayTicks = this.title.toTicks(times.stay());
            int outTicks = this.title.toTicks(times.fadeOut());
            if (inTicks != -1 || stayTicks != -1 || outTicks != -1) {
               this.title.contributeTimes(collection, inTicks, stayTicks, outTicks);
            }
         }

         title = this.title.completeTitle(collection);
         if (title != null) {
            Iterator var9 = this.viewers.iterator();

            while(var9.hasNext()) {
               V viewer = var9.next();
               this.title.showTitle(viewer, title);
            }

         }
      }
   }

   public void clearTitle() {
      if (this.title != null) {
         Iterator var1 = this.viewers.iterator();

         while(var1.hasNext()) {
            V viewer = var1.next();
            this.title.clearTitle(viewer);
         }

      }
   }

   public void resetTitle() {
      if (this.title != null) {
         Iterator var1 = this.viewers.iterator();

         while(var1.hasNext()) {
            V viewer = var1.next();
            this.title.resetTitle(viewer);
         }

      }
   }

   public void showBossBar(final @NotNull BossBar bar) {
      if (this.bossBar != null && this.bossBars != null) {
         Object listener;
         synchronized(this.bossBars) {
            listener = (Facet.BossBar)this.bossBars.get(bar);
            if (listener == null) {
               listener = new FacetBossBarListener(this.bossBar.createBossBar(this.viewers), (message) -> {
                  return this.provider.componentRenderer.render(message, this);
               });
               this.bossBars.put(bar, listener);
            }
         }

         if (((Facet.BossBar)listener).isEmpty()) {
            ((Facet.BossBar)listener).bossBarInitialized(bar);
            bar.addListener((BossBar.Listener)listener);
         }

         Iterator var3 = this.viewers.iterator();

         while(var3.hasNext()) {
            V viewer = var3.next();
            ((Facet.BossBar)listener).addViewer(viewer);
         }

      }
   }

   public void hideBossBar(final @NotNull BossBar bar) {
      if (this.bossBars != null) {
         Facet.BossBar<V> listener = (Facet.BossBar)this.bossBars.get(bar);
         if (listener != null) {
            Iterator var3 = this.viewers.iterator();

            while(var3.hasNext()) {
               V viewer = var3.next();
               listener.removeViewer(viewer);
            }

            if (listener.isEmpty() && this.bossBars.remove(bar) != null) {
               bar.removeListener(listener);
               listener.close();
            }

         }
      }
   }

   public void sendPlayerListHeader(final @NotNull Component header) {
      if (this.tabList != null) {
         Object headerFormatted = this.createMessage(header, this.tabList);
         if (headerFormatted == null) {
            return;
         }

         Iterator var3 = this.viewers.iterator();

         while(var3.hasNext()) {
            V viewer = var3.next();
            this.tabList.send(viewer, headerFormatted, (Object)null);
         }
      }

   }

   public void sendPlayerListFooter(final @NotNull Component footer) {
      if (this.tabList != null) {
         Object footerFormatted = this.createMessage(footer, this.tabList);
         if (footerFormatted == null) {
            return;
         }

         Iterator var3 = this.viewers.iterator();

         while(var3.hasNext()) {
            V viewer = var3.next();
            this.tabList.send(viewer, (Object)null, footerFormatted);
         }
      }

   }

   public void sendPlayerListHeaderAndFooter(final @NotNull Component header, final @NotNull Component footer) {
      if (this.tabList != null) {
         Object headerFormatted = this.createMessage(header, this.tabList);
         Object footerFormatted = this.createMessage(footer, this.tabList);
         if (headerFormatted == null || footerFormatted == null) {
            return;
         }

         Iterator var5 = this.viewers.iterator();

         while(var5.hasNext()) {
            V viewer = var5.next();
            this.tabList.send(viewer, headerFormatted, footerFormatted);
         }
      }

   }

   public @NotNull Pointers pointers() {
      if (this.pointers == null) {
         synchronized(this) {
            if (this.pointers == null) {
               V viewer = this.viewer;
               if (viewer == null) {
                  return Pointers.empty();
               }

               Pointers.Builder builder = Pointers.builder();
               this.contributePointers(builder);
               Iterator var4 = this.pointerProviders.iterator();

               while(var4.hasNext()) {
                  Facet.Pointers<V> provider = (Facet.Pointers)var4.next();
                  if (provider.isApplicable(viewer)) {
                     provider.contributePointers(viewer, builder);
                  }
               }

               return this.pointers = (Pointers)builder.build();
            }
         }
      }

      return this.pointers;
   }

   @OverrideOnly
   protected void contributePointers(final Pointers.Builder builder) {
   }

   public void close() {
      Iterator var1;
      if (this.bossBars != null) {
         var1 = (new LinkedList(this.bossBars.keySet())).iterator();

         while(var1.hasNext()) {
            BossBar bar = (BossBar)var1.next();
            this.hideBossBar(bar);
         }

         this.bossBars.clear();
      }

      var1 = this.viewers.iterator();

      while(var1.hasNext()) {
         V viewer = var1.next();
         this.removeViewer(viewer);
      }

      this.viewers.clear();
   }

   private @Nullable Object createMessage(final @NotNull Component original, final Facet.@NotNull Message<V, Object> facet) {
      Component message = this.provider.componentRenderer.render(original, this);
      V viewer = this.viewer;
      return viewer == null ? null : facet.createMessage(viewer, message);
   }
}
